//	C++ Include
#include <iostream>
#include <vector>
#include <set>

//	Dune Includes
#include "config.h"

//	Project Specific Dune Includes 
#include <dune/grid/yaspgrid.hh>					//	functions to create yasp grid
#include <dune/grid/common/gridinfo.hh>				//	functions to provide information about the grid
#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/bvector.hh>
#include <dune/istl/io.hh>
#include <dune/grid/common/quadraturerules.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/ilu.hh>
#include <dune/common/stdstreams.hh>
#include <dune/common/debugstream.hh>
#include <dune/grid/io/file/vtk/vtkwriter.hh>
#include <dune/istl/operators.hh> 
#include <dune/istl/solvers.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/io.hh>


//#include <dune/grid/common/genericreferenceelements.hh>
//	Project Specific Custom Includes
#include "shapefunctions.hh"


template<class GV, class F>
class  P1Elements{
	public:
		static const int dim = GV::dimension;
		
		typedef typename GV::ctype ctype;
		typedef Dune::BCRSMatrix<Dune::FieldMatrix<ctype, 1, 1> > Matrix;
		typedef Dune::BlockVector<Dune::FieldVector<ctype, 1> > ScalarField;

	private:
		typedef typename GV::template Codim<0>::Iterator LeafIterator;
		typedef typename GV::IntersectionIterator IntersectionIterator;
		typedef typename GV::IndexSet LeafIndexSet;

		const GV& gv;
		const F& f;

	public:
		Matrix A;
		ScalarField b;
		ScalarField u;
		
		std::vector<std::set<int> > adjacencyPattern;
		
		P1Elements(const GV& gv_, const F& f_) : gv(gv_), f(f_) {}		
	
		void determineAdjacencyPattern();
		
		void assemble();
		void solve();

};


template<class GV, class F>
void P1Elements<GV, F>::determineAdjacencyPattern(){
	const int N = gv.size(dim);
	adjacencyPattern.resize(N);
	
	const LeafIndexSet& set = gv.indexSet();
	const LeafIterator itend = gv.template end<0>();
	
	for(LeafIterator it = gv.template begin<0>(); it!=itend; ++it){
		Dune::GeometryType gt =it->type();
		const Dune::template GenericReferenceElement<ctype,dim> &ref =
                Dune::GenericReferenceElements<ctype,dim>::general(gt);

		
		const IntersectionIterator isend = gv.iend(*it);
		
		for(IntersectionIterator is=gv.ibegin(*it); is!=isend ; ++is ){
			int vertexsize = ref.size(is->indexInInside(),1,dim);
			for(int i=0; i<vertexsize; ++i){
				int indexi = set.subIndex(*it,ref.subEntity(is->indexInInside(),1,i,dim),dim);
				for(int j=0; j<vertexsize; ++j){
					int indexj = set.subIndex(*it, ref.subEntity(is->indexInInside(),1,j,dim),dim);
					adjacencyPattern[indexi].insert(indexj);
				}
			}
		}
	}
}

template<class GV, class F>
void P1Elements<GV, F>::assemble()
{
    const int N = gv.size(dim);

    const LeafIndexSet& set = gv.indexSet();
    const LeafIterator itend = gv.template end<0>();

    // set sizes of A and b
    A.setSize(N, N, N + 2*gv.size(dim-1));
    A.setBuildMode(Matrix::random);
    b.resize(N, false);

    for (int i = 0; i < N; i++)
        A.setrowsize(i,adjacencyPattern[i].size());
    A.endrowsizes();

    // set sparsity pattern of A with the information gained in determineAdjacencyPattern 
    for (int i = 0; i < N; i++) /*@\label{fem:setpattern}@*/
    {
        std::template set<int>::iterator setend = adjacencyPattern[i].end();
        for (std::template set<int>::iterator setit = adjacencyPattern[i].begin();
             setit != setend; ++setit)
            A.addindex(i,*setit);
    }

    A.endindices(); /*@\label{fem:endindices}@*/

    // initialize A and b
    A = 0.0;
    b = 0.0;

    // get a set of P1 shape functions
    P1ShapeFunctionSet<ctype,ctype,dim> basis = P1ShapeFunctionSet<ctype,ctype,dim>::instance();

    for (LeafIterator it = gv.template begin<0>(); it != itend; ++it) /*@\label{fem:loop1}@*/
    {
        // determine geometry type of the current element and get the matching reference element
        Dune::GeometryType gt = it->type();
        const Dune::template GenericReferenceElement<ctype,dim> &ref =
            Dune::GenericReferenceElements<ctype,dim>::general(gt);
        int vertexsize = ref.size(dim);

        // get a quadrature rule of order one for the given geometry type
        const Dune::QuadratureRule<ctype,dim>& rule = Dune::QuadratureRules<ctype,dim>::rule(gt,1);
        for (typename Dune::QuadratureRule<ctype,dim>::const_iterator r = rule.begin();
             r != rule.end() ; ++r)
        {
            // compute the jacobian inverse transposed to transform the gradients
            Dune::FieldMatrix<ctype,dim,dim> jacInvTra =
                it->geometry().jacobianInverseTransposed(r->position());

            // get the weight at the current quadrature point
            ctype weight = r->weight();

            // compute Jacobian determinant for the transformation formula
            ctype detjac = it->geometry().integrationElement(r->position());
            for (int i = 0; i < vertexsize; i++)
            {
                // compute transformed gradients
                Dune::FieldVector<ctype,dim> grad1;
                jacInvTra.mv(basis[i].evaluateGradient(r->position()),grad1);
                for (int j = 0; j < vertexsize; j++)
                {
                    Dune::FieldVector<ctype,dim> grad2;
                    jacInvTra.mv(basis[j].evaluateGradient(r->position()),grad2);

                    // gain global inidices of vertices i and j and update associated matrix entry
                    A[set.subIndex(*it,i,dim)][set.subIndex(*it,j,dim)] /*@\label{fem:calca}@*/
                        += (grad1*grad2) * weight * detjac; 
                }
            }
        }

        // get a quadrature rule of order two for the given geometry type
        const Dune::QuadratureRule<ctype,dim>& rule2 = Dune::QuadratureRules<ctype,dim>::rule(gt,2);
        for (typename Dune::QuadratureRule<ctype,dim>::const_iterator r = rule2.begin();
             r != rule2.end() ; ++r)
        {
            ctype weight = r->weight();
            ctype detjac = it->geometry().integrationElement(r->position());
            for (int i = 0 ; i<vertexsize; i++)
            {
                // evaluate the integrand of the right side
                ctype fval = basis[i].evaluateFunction(it->geometry().global(r->position()))
                    * f(it->geometry().global(r->position())) ;
                b[set.subIndex(*it,i,dim)] += fval * weight * detjac; /*@\label{fem:calcb}@*/
            }
        }
    } /*@\label{fem:loop2}@*/

    // Dirichlet boundary conditions:
    // replace lines in A related to Dirichlet vertices by trivial lines
    for ( LeafIterator it = gv.template begin<0>() ; it != itend ; ++it) /*@\label{fem:boundary1}@*/
    {
        const IntersectionIterator isend = gv.iend(*it);
        for (IntersectionIterator is = gv.ibegin(*it) ; is != isend ; ++is)
        {
            // determine geometry type of the current element and get the matching reference element
            Dune::GeometryType gt = it->type();
            const Dune::template GenericReferenceElement<ctype,dim> &ref =
                Dune::GenericReferenceElements<ctype,dim>::general(gt);

            // check whether current intersection is on the boundary
            if ( is->boundary() )
            {
                // traverse all vertices the intersection consists of
                for (int i=0; i < ref.size(is->indexInInside(),1,dim); i++)
                {
                    // and replace the associated line of A and b with a trivial one
                    int indexi = set.subIndex(*it,ref.subEntity(is->indexInInside(),1,i,dim),dim);

                    A[indexi] = 0.0; /*@\label{fem:trivialline}@*/
                    A[indexi][indexi] = 1.0;
                    b[indexi] = 0.0;
                }
            }
        }
    } /*@\label{fem:boundary2}@*/
}

template<class GV, class E>
void P1Elements<GV, E>::solve()
{
    // make linear operator from A
    Dune::MatrixAdapter<Matrix,ScalarField,ScalarField> op(A);

    // initialize preconditioner
    Dune::SeqILUn<Matrix,ScalarField,ScalarField> ilu1(A, 1, 0.92);

    // the inverse operator
    Dune::BiCGSTABSolver<ScalarField> bcgs(op, ilu1, 1e-15, 5000, 0);
    Dune::InverseOperatorResult r;

    // initialue u to some arbitrary value to avoid u being the exact
    // solution
    u.resize(b.N(), false);
    u = 2.0;

    // finally solve the system
    bcgs.apply(u, b, r);
}


template <class ctype, int dim>
class Bump{
	public:
		ctype operator() (Dune::FieldVector<ctype,dim> x) const{
			ctype result=0;
			for(int i=0; i<dim; ++i)
				result += 2.0*x[i]*(1-x[i]);
			return result;
		}
};

int main(){
	
	//a try block to catch error messages from dune
	try{
		const int dim =2;																		//	specify the dimension of the grid
		typedef Dune::YaspGrid<dim> GridType;													//	specify the grid to be used
		typedef GridType::ctype		ctype;														//	extract the type used by the grids to store the co-ordinates
		typedef GridType::LeafGridView GV;														//  extract the grdiview type

		typedef Bump<ctype,dim> Func;															//	define which function to use
		
		//creating the grid
		Dune::FieldVector<int,dim>		N(10);													//	specify the number of elements created in all directions
		Dune::FieldVector<ctype,dim>	U(1);													//	specify the top right corner of the grid
																								// 	YaspGrid hardwires the bottom right corner to zero

		Dune::FieldVector<bool,dim>		periodic(false);										//	specifiy the peridocity of the co-ordinates
		int load_balancing=0;																	//	disable load balancing	(YaspGrid does not support Load Balacing)
		
		GridType grid(U,N,periodic,load_balancing);												//	create the yasp grid with the parameters 
																								//	U	->	upper corner co-ordinates
																								//	N	->	number of elements in direction of axes
																								//	periodic	->	peridocity of the co-ordinates
																								//	load_balancing	->	 to support load balancing or not 
		//Dune::gridinfo(grid);
		//Dune::gridlevellist(grid,"\t",0);



		const GV& gv=grid.leafView();															// define which gridview to use

		Func f;																					// instantiate the function
		
		

		P1Elements<GV,Func> p1(gv,f);
		grid.globalRefine(4);
		std::cout<<" Number of Unknowns "<<grid.size(dim)<<std::endl; 
		p1.determineAdjacencyPattern();
		//std::cout<<adjacencyPattern<<std::endl;
		p1.assemble();
		
	}
	catch (std::exception & e) {
    	std::cout << "STL ERROR: " << e.what() << std::endl;
    	return 1;
	}
  	catch (Dune::Exception & e) {
    	std::cout << "DUNE ERROR: " << e.what() << std::endl;
    	return 1;
	}
	catch (...) {
    	std::cout << "Unknown ERROR" << std::endl;
    	return 1;
	}
	
	//program complete successfully
	return 0;
}


